/*
 * Grid.cpp
 *
 *  Created on: Sep 11, 2012
 *      Author: scottyoung
 */

#include "Grid.h"

#define EPSILON .0001

Grid::Grid(Vec3f lo, Vec3f hi, int x, int y, int z, Material *mat)
	{
		//SET THIS FLAG TO TRUE IF YOU WANT TO SEE THE GRID
		RENDER_GRID = false;
		//------------------
		min = lo; nx=x;
		max = hi; ny=y;
		nz = z; m = mat;
		dim = Vec3f((hi.x()-lo.x())/nx, (hi.y()-lo.y())/ny, (hi.z()-lo.z())/nz);
		cells.resize(nx);
		objects.resize(nx);
		for(int i=0; i<x; i++)
		{
			cells[i].resize(ny);
			objects[i].resize(ny);
			for(int j=0; j<y; j++)
			{
				cells[i][j].resize(nz);
				objects[i][j].resize(nz);
				for(int k=0; k<z; k++)
				{
					objects[i][j][k] = new Object3DVector();
					cells[i][j][k] = false;
				}
			}
		}
	}

Grid::~Grid() {
	// TODO Auto-generated destructor stub
}

void Grid::rasterizeSphere(Vec3f center, float radius) {
	Vec3f start = min + (dim*.5);
	for(int x=0; x<nx; x++)
	{
		for(int y=0; y<ny; y++)
		{
			for(int z=0; z<nz; z++)
			{
				Vec3f point(start.x()+(dim.x()*x), start.y()+(dim.y()*y), start.z()+(dim.z()*z));
				float sqrdDist = (point-center).Dot3(point-center);
				if(sqrdDist <= radius*radius)
				{
					cells[x][y][z] = true;
				}
			}
		}
	}
}

void Grid::setBounds(){
	bounds = BoundingBox(min, max);
}

bool Grid::intersect(const Ray& r, Hit& h, float tMin) {

	MarchingInfo march = initializeMarchingRay(r,tMin);
	int x = march.getX();
	int y = march.getY();
	int z = march.getZ();
	bool hit = false;
	while(x<nx && x>=0 && y<ny && y>=0 && z<nz && z>=0)
	{
		if(RENDER_GRID && cells[x][y][z]) //If we want to render only the grid
		{
			h.set(march.getT(), m, march.getN(), r);
			return(true);
		}
		else if(objects[x][y][z]->getNumObjects() > 0) //If we want to render objects *in* the grid
		{
			Object3DVector* objs = objects[x][y][z];
			int n = objs->getNumObjects();
			bool InnerHit = false;
			for(int i=0; i<n; i++)
			{
				if(objs->getObject(i)->intersect(r, h, tMin))
				{
					Vec3f lo(min.x()+dim.x()*x, min.y()+dim.y()*y, min.z()+dim.z()*z);
					Vec3f hi(min.x()+dim.x()*(x+1), min.y()+dim.y()*(y+1), min.z()+dim.z()*(z+1));
					BoundingBox b = BoundingBox(Vec3f(lo), Vec3f(hi));
					//Only exit the loop if we have intersected with an object in *this* cell, otherwise
					//there may be one in another cell which is closer
					if(b.isInside(h.getIntersectionPoint()))
					{
						InnerHit = true;
					}
					hit = true; //Mark this as true so the function knows that *something* was hit, even if it was outside this cell
				}
			}
			//If we hit an object exit within this cell, no need to continue march
			if(InnerHit)
				return true;
		}
		march.nextCell();
		x = march.getX();
		y = march.getY();
		z = march.getZ();
	}
	return(hit);
}

void Grid::printContent() {
	for(int i=0; i<nx; i++)
	{
		cout << endl << "XSlice: " << i << endl ;
		for(int j=0; j<ny; j++)
		{
			for(int k=0; k<nz; k++)
			{
				cout << cells[i][j][k] << " ";
			}
			cout << endl;
		}
	}
}

/**
 * Makes a grid, given a group.
 */
Grid* Grid::make(Group *theGroup, int nx, int ny, int nz){

	theGroup->setBounds();
	BoundingBox b = theGroup->getBoundingBox();

	Vec3f wiggle(EPSILON, EPSILON, EPSILON);

	Material *m = (Material*) new PhongMaterial(Vec3f(1,1,1));
	Grid *theGrid = new Grid(b.getMin()-wiggle, b.getMax()+wiggle, nx, ny, nz, m);

	theGrid->insertGroup(theGroup);

	return(theGrid);
}

void Grid::insertGroup(Group *g){

	Object3D* obj = g->firstObject();
	while(obj != 0)
	{
		if(obj->isGroup)
			insertGroup((Group*)obj);
		else
			insertObject(obj);
		obj = g->nextObject();
	}
}

/**
 * Inserts an individual object into a grid's appropriate squares
 */
void Grid::insertObject(Object3D* obj){

	BoundingBox b = obj->getBoundingBox();
	Vec3f lo = b.getMin();
	Vec3f hi = b.getMax();

	int x_lo = (lo.x()-min.x())/dim.x();
	int y_lo = (lo.y()-min.y())/dim.y();
	int z_lo = (lo.z()-min.z())/dim.z();

	int x_hi = (hi.x()-min.x())/dim.x();
	int y_hi = (hi.y()-min.y())/dim.y();
	int z_hi = (hi.z()-min.z())/dim.z();

	//UNCOMMENT TO GET INSERTION STATS
	/*cout << "INSERTION STATS: " << endl; //DEBUG
	cout << "[" << x_lo << "," << y_lo << "," << z_lo << "] -> [" << x_hi << "," << y_hi << "," << z_hi << "]";
	cout << "lo: "; lo.Write();
	cout << "hi: "; hi.Write();
	cout << "min:"; min.Write();
	cout << "dim:"; dim.Write();*/

	if(x_lo < 0) x_lo=0;
	if(y_lo < 0) y_lo=0;
	if(z_lo < 0) z_lo=0;
	if(x_hi >= nx) x_hi = nx-1;
	if(y_hi >= ny) y_hi = ny-1;
	if(z_hi >= nz) z_hi = nz-1;

	for(int x = x_lo; x <= x_hi; x++){
		for(int y = y_lo; y <= y_hi; y++){
			for(int z = z_lo; z <= z_hi; z++){
				objects[x][y][z]->addObject(obj);
				cells[x][y][z] = true;
			}
		}
	}

}

