/*
 * MarchingInfo.cpp
 *
 *  Created on: Sep 11, 2012
 *      Author: scottyoung
 */

#include "MarchingInfo.h"
#define EPSILON .0001

MarchingInfo::MarchingInfo(Vec3f min, Vec3f max, int nx, int ny, int nz, const Ray& r, float tMinInitial) {

	Vec3f dir = r.getDirection();
	Vec3f org = r.getOrigin();

	if(dir.x() < 0) signX = -1; else signX = 1;
	if(dir.y() < 0) signY = -1; else signY = 1;
	if(dir.z() < 0) signZ = -1; else signZ = 1;

	tMin = tMinInitial;

	Vec3f start = r.pointAtParameter(tMinInitial);
	Vec3f grid((max.x()-min.x())/nx, (max.y()-min.y())/ny, (max.z()-min.z())/nz);

	normal = Vec3f(1,1,1);

	//Initialize the spacing between hits
	dt = Vec3f(fabs(grid.x()/dir.x()), fabs(grid.y()/dir.y()), fabs(grid.z()/dir.z()));

	//CASE 1: We're inside the grid
	if(start.insideBoundingBox(min, max)){

		currX = (start.x()-min.x())/grid.x();
		currY = (start.y()-min.y())/grid.y();
		currZ = (start.z()-min.z())/grid.z();

		tNext = calculateTNext(grid, min, start, dir);


	} else {
	//CASE 2: We're outside the grid
		BoundingBox b(min,max);
		float tVal;
		if(b.intersect(r, tVal) && tVal >= tMin)
		{
			tMin = tVal;
			Vec3f point = r.pointAtParameter(tVal);

			currX = (point.x()-min.x())/grid.x();
			currY = (point.y()-min.y())/grid.y();
			currZ = (point.z()-min.z())/grid.z();

			//Make sure we're entering a cell which exists
			if(currX==nx) currX--;
			if(currY==ny) currY--;
			if(currZ==nz) currZ--;


			tNext = calculateTNext(grid, min, point, dir);

			//Establish the normal
			if(fabs(point.x()-min.x()) < EPSILON)
				normal = Vec3f(-1,0,0);
			else if(fabs(point.x()-max.x()) < EPSILON)
				normal = Vec3f(1,0,0);
			else if(fabs(point.y()-min.y()) < EPSILON)
				normal = Vec3f(0,-1,0);
			else if(fabs(point.y()-max.y()) < EPSILON)
				normal = Vec3f(0,1,0);
			else if(fabs(point.z()-min.z()) < EPSILON)
				normal = Vec3f(0,0,-1);
			else if(fabs(point.z()-max.z()) < EPSILON)
				normal = Vec3f(0,0,1);

			//cout << "Dot intersects with: " << currX << "," << currY << "," << currZ << endl;//DEBUG

		}else{
			//We never touch the box, set current cell to -1,-1,-1
			currX = -1; currY = -1; currZ = -1;
		}
	}

	RayTracingStats::IncrementNumGridCellsTraversed();
	t = tMin;
}

MarchingInfo::~MarchingInfo() {
	// TODO Auto-generated destructor stub
}

void MarchingInfo::nextCell() {
	if(fabs(tNext.x()-tNext.Min()) < EPSILON)
	{
		currX += signX;
		t = tMin + tNext.x();
		tNext.Set(tNext.x()+dt.x(), tNext.y(), tNext.z());
		normal = Vec3f(-signX,0,0);
	}else if(fabs(tNext.y()-tNext.Min()) < EPSILON){
		currY += signY;
		t = tMin + tNext.y();
		tNext.Set(tNext.x(), tNext.y()+dt.y(), tNext.z());
		normal = Vec3f(0,-signY,0);
	}else{
		currZ += signZ;
		t = tMin + tNext.z();
		tNext.Set(tNext.x(), tNext.y(), tNext.z()+dt.z());
		normal = Vec3f(0,0,-signZ);
	}
}

Vec3f MarchingInfo::calculateTNext(Vec3f grid, Vec3f min, Vec3f point, Vec3f dir){
	//If the sign is positive we want to intersect with the beginning of the
	//*next* cube. If the sign is negative, we want to intersect with the beginning
	//of *this* cube.
	float tNextX, tNextY, tNextZ;

	RayTracingStats::IncrementNumGridCellsTraversed();

	if(signX>0)
		tNextX = ((currX+1)*grid.x()+min.x()-point.x())/dir.x();
	else
		tNextX = (currX*grid.x()+min.x()-point.x())/dir.x();
	if(signY>0)
		tNextY = ((currY+1)*grid.y()+min.y()-point.y())/dir.y();
	else
		tNextY = (currY*grid.y()+min.y()-point.y())/dir.y();
	if(signZ>0)
		tNextZ = ((currZ+1)*grid.z()+min.z()-point.z())/dir.z();
	else
		tNextZ = (currZ*grid.z()+min.z()-point.z())/dir.z();

	return(Vec3f(tNextX, tNextY, tNextZ));
}
