/*
 * RayTracer.cpp
 *
 *  Created on: Sep 4, 2012
 *      Author: scottyoung
 */

#include "RayTracer.h"

RayTracer::~RayTracer() {
	// TODO Auto-generated destructor stub
}

Vec3f RayTracer::traceRay(const Ray &r, float tMin, int bounces, float weight, float indexOfRefraction, Hit &h, int DEBUG_CODE) const{
	int nL = theScene->getNumLights();
	Group *group = theScene->getGroup();
	Light **lights = theScene->getLights();
	Vec3f ambient = theScene->getAmbientLight();
	Material *m = h.getMaterial();
	Vec3f value(m->flatColor());
	value.Scale(ambient.r(), ambient.g(), ambient.b());


	//First, calculate contribution from lights
	for(int i=0; i<nL; i++)
	{
		bool shadow = false;
		RayTracingStats::IncrementNumShadowRays();

		Vec3f I, dir;
		float distance;
		lights[i]->getIllumination(h.getIntersectionPoint(), dir, I, distance);
		Vec3f epsilon = h.getIntersectionPoint();
		Ray shadowRay(dir, epsilon);
		Hit shadowHit;
		if(castShadows)
			shadow = group->intersect(shadowRay, shadowHit, EPSILON);

		//Do a quick check to see how far we are from the point light, to make sure it's not between us and the object
		float distToShadowHit = (shadowHit.getIntersectionPoint()-shadowRay.getOrigin()).Length();
		if(distToShadowHit > distance)
			shadow = false;

		if(!shadow)
		{
			Vec3f phong = m->shade(r,&h,dir,I);
			value += phong;
		}
	}


	if(bounces < maxBounces && weight > minWeight)
	{
		Vec3f E(r.getDirection());
		E.Negate();
		float ior = indexOfRefraction/m->getIndex();
		Vec3f N(h.getNormal());

		//Check if we are coming from inside to outside the object, in which case flip the normal
		//and the index becomes the reciprocal (opposite direction of outside/inside)
		if(E.Dot3(N) < 0)
		{
			N.Negate();
			ior = 1/ior;
		}
		//Now calculate contribution from reflection
		if(m->reflectWeight() > minWeight)
		{
			Vec3f reflectDir = Mirror(E, N), reflection;
			Ray reflectRay(reflectDir, h.getIntersectionPoint());
			Hit reflectHit;

			RayTracingStats::IncrementNumNonShadowRays();

			if(group->intersect(reflectRay, reflectHit, EPSILON))
			{
				//This computes weight as an average of the RGB weights for the reflection vector
				float newWeight = m->reflectWeight();
				reflection = traceRay(reflectRay, EPSILON, bounces+1, weight*newWeight, indexOfRefraction, reflectHit, 1);
			}
			else //Just return the background color w/ no intersections
			{
				reflection = theScene->getBackgroundColor();
			}
			Vec3f weights = m->getReflectiveColor();
			reflection.Scale(weights.r(), weights.g(), weights.b());
			value += reflection;
		}


		//Now calculate contribution from transparency
		if(m->transparency() > minWeight)
		{
			Vec3f refractDir = transmission(E, N, ior);
			//Don't bother if there was total internal reflection
			if(refractDir.Length() > .001)
			{
				Vec3f refraction;
				Ray refractRay(refractDir, h.getIntersectionPoint());
				Hit refractHit;

				RayTracingStats::IncrementNumNonShadowRays();

				if(group->intersect(refractRay, refractHit, EPSILON))
				{
					float newWeight = m->transparency();
					refraction = traceRay(refractRay, EPSILON, bounces+1, weight*newWeight, m->getIndex(), refractHit, 2);
				}
				else
				{
					refraction = theScene->getBackgroundColor();
				}
				Vec3f weights = m->getTransparentColor();
				refraction.Scale(weights.r(), weights.g(), weights.b());
				value += refraction;
			}
		}
	}


	return(value);
}

Vec3f RayTracer::transmission(const Vec3f &I, const Vec3f &N, float ratioOfIndices)
{
	float ratiosSquared = ratioOfIndices*ratioOfIndices;
	float IdotN = I.Dot3(N);
	float adjustment = 1 - ratiosSquared*(1-IdotN*IdotN);
	Vec3f transmitted(0,0,0);
	//Calculate if we do not have total internal reflection
	if(adjustment >= 0)
	{
		transmitted = N*(ratioOfIndices*IdotN-sqrt(adjustment)) - I*ratioOfIndices;
		transmitted.Normalize();
	}
	return(transmitted);
}
