/*
 * Sphere.cpp
 *
 *  Created on: Aug 10, 2012
 *      Author: scottyoung
 */

#include "Sphere.h"
#include <math.h>

#define PI 3.14159265359
#define N_THETA 50
#define N_PHI 25
#define GOURAUD true

Sphere::Sphere(Vec3f c, float r, Material *m) {
	center = c;
	radius = r;
	material = m;
	isGroup=false;
	isTri=false;
}

Sphere::~Sphere() {

}

bool Sphere::intersect(const Ray &r, Hit &h, float tMin)
{
	bool hit = false;

	Vec3f E = r.getOrigin();
	Vec3f R = E - center;
	Vec3f D = r.getDirection();
	Vec3f N;


	RayTracingStats::IncrementNumIntersections();


	float a = D.Dot3(D); //in transformed objects, dir may not be normalized
	float b = R.Dot3(D);
	float c = R.x()*R.x() + R.y()*R.y() + R.z()*R.z() - radius*radius;

	float discriminant = b*b-a*c;
	float hit1, hit2, tValue;

	/*cout << endl << "Testing Sphere Intersection " << endl; //DEBUG
	cout << "Discriminant: " << discriminant << endl;*/

	if(discriminant >= 0)
	{
		//Find the closest intersection
		hit1 = (-b  +  sqrt(discriminant)) / a;
		hit2 = (-b  -  sqrt(discriminant)) / a;

		if(hit2 > tMin && hit1 > tMin)
			tValue = min(hit1, hit2);
		else if(hit2 > tMin)
			tValue = hit2;
		else
			tValue = hit1;

		/*cout << "Hit 1: " << hit1 << endl;//DEBUG
		cout << "Hit 2: " << hit2 << endl;
		cout << "TVal:  " << tValue << endl;
		cout << "TMin:  " << tMin << endl;
		cout << "Current Best T: " << h.getT() << endl << endl;*/

		if(tValue > tMin && tValue < h.getT())
		{
			//Find the normal
			N = R+D*tValue;
			N.Normalize();
			h.set(tValue, material, N, r);
			hit = true;
		}
	}

	return(hit);
}

void Sphere::setBounds(){
	Vec3f lo(center.x()-radius, center.y()-radius, center.z()-radius);
	Vec3f hi(center.x()+radius, center.y()+radius, center.z()+radius);
	bounds = BoundingBox(lo,hi);
}

void Sphere::print(){
	cout << "Sphere: pos @(" << center.x() <<","<< center.y()<<","<<center.z()<<") radius = "<< radius << endl;
}

/**
 * paints the sphere, checks whether global variable GOURAUD = true and uses N_THETA and N_PHI to determine
 * the tesselation parameters
*/
void Sphere::paint(){
	material->glSetMaterial();
	glBegin(GL_TRIANGLES);
		for(int iPhi = 0; iPhi < N_PHI; iPhi++)
		{
			for(int iTheta = 0; iTheta < N_THETA; iTheta++)
			{
				bool oneLow = false, oneHigh = false;
				float phi_lo = (float)(iPhi)/N_PHI * PI - PI/2;
				float phi_hi = (float)(iPhi+1)/N_PHI * PI - PI/2;
				//If we're at the top don't create squares, create triangles (i.e. there is only one theta @ the high/low position)
				if(iPhi == 0)
					oneLow = true;
				else if(iPhi == N_PHI-1)
					oneHigh = true;

				float theta_lo = (float)(iTheta)/N_THETA * 2 * PI;
				float theta_hi = (float)(iTheta+1)/N_THETA * 2 * PI;

				Vec3f top_right;
				Vec3f top_left;
				Vec3f bottom_right;
				Vec3f bottom_left;
				Vec3f norm;
				Vec3f tr_norm, tl_norm, br_norm, bl_norm;

				//Calculate the vertices for the unit sphere
				top_right.Set(cos(theta_hi)*cos(phi_hi), sin(phi_hi), sin(theta_hi)*cos(phi_hi));
				top_left.Set(cos(theta_lo)*cos(phi_hi), sin(phi_hi), sin(theta_lo)*cos(phi_hi));
				bottom_right.Set(cos(theta_hi)*cos(phi_lo), sin(phi_lo), sin(theta_hi)*cos(phi_lo));
				bottom_left.Set(cos(theta_lo)*cos(phi_lo), sin(phi_lo), sin(theta_lo)*cos(phi_lo));

				//Now scale these vertices to the radius
				top_right *= radius;
				top_left *= radius;
				bottom_right *= radius;
				bottom_left *= radius;

				//Get the normals of each vertex for use in Gouraud interpolation
				tr_norm = top_right; tr_norm.Normalize();
				tl_norm = top_left; tl_norm.Normalize();
				br_norm = bottom_right; br_norm.Normalize();
				bl_norm = bottom_left; bl_norm.Normalize();

				//Now translate them to the center
				top_right += center;
				top_left += center;
				bottom_right += center;
				bottom_left += center;

				//If we aren't using Gouraud interpolation, just set the normal to the triangle normal
				if(!oneHigh)
				{
					if(!GOURAUD)
					{
						Vec3f::Cross3(norm, (top_right-top_left), (top_right-bottom_right));
						glNormal3f(norm.x(), norm.y(), norm.z());
					}
					else
						glNormal3f(tr_norm.x(), tr_norm.y(), tr_norm.z());
					glVertex3f(top_right.x(), top_right.y(), top_right.z());
					if(GOURAUD)
						glNormal3f(br_norm.x(), br_norm.y(), br_norm.z());
					glVertex3f(bottom_right.x(), bottom_right.y(), bottom_right.z());
					if(GOURAUD)
						glNormal3f(tl_norm.x(), tl_norm.y(), tl_norm.z());
					glVertex3f(top_left.x(), top_left.y(), top_left.z());
				}
				//Now complete the second triangle
				if(!oneLow)
				{
					if(!GOURAUD)
					{
						Vec3f::Cross3(norm, (bottom_right-bottom_left), (bottom_right-top_left));
						glNormal3f(norm.x(), norm.y(), norm.z());
					}
					else
						glNormal3f(br_norm.x(), br_norm.y(), br_norm.z());
					glVertex3f(bottom_right.x(), bottom_right.y(), bottom_right.z());
					if(GOURAUD)
						glNormal3f(bl_norm.x(), bl_norm.y(), bl_norm.z());
					glVertex3f(bottom_left.x(), bottom_left.y(), bottom_left.z());
					if(GOURAUD)
						glNormal3f(tl_norm.x(), tl_norm.y(), tl_norm.z());
					glVertex3f(top_left.x(), top_left.y(), top_left.z());
				}

			}
		}

	glEnd();
}
