#ifndef _LIGHT_H_
#define _LIGHT_H_

#include "vectors.h"

// ====================================================================
// ====================================================================

class Light {

public:

  // CONSTRUCTOR & DESTRUCTOR
  Light() {}
  virtual ~Light() {}

  // VIRTUAL METHOD
  virtual void getIllumination (const Vec3f &p, Vec3f &dir, Vec3f &col, float &distance) const = 0;
  virtual void print()=0;
  virtual void glInit(int i)=0;

protected:
  Vec3f color;

};

// ====================================================================
// ====================================================================

class DirectionalLight : public Light {

public:

  // CONSTRUCTOR & DESTRUCTOR
  DirectionalLight() { 
    direction = Vec3f(0,0,0);
    color = Vec3f(1,1,1); }
  DirectionalLight(const Vec3f &d, const Vec3f &c) {
    direction = d; direction.Normalize();
    color = c; }
  virtual ~DirectionalLight() {}

  // VIRTUAL METHOD
  void getIllumination (const Vec3f &p, Vec3f &dir, Vec3f &col, float &distance) const {
    // the direction to the light is the opposite of the
    // direction of the directional light source
    dir = direction * (-1.0f);
    distance = INFINITY;
    col = color; }

  void print(){
	  cout << "Directional Light: dir=(" << direction.x() << "," << direction.y() << "," << direction.z() << ")"
			<< " color=(" << color.r() << "," << color.g() << "," << color.b() << ")" << endl;
  }

  // ====================================================================
  // Create an OpenGL directional light source with the appropriate
  // color and position
  // ====================================================================

  void glInit(int id) {
	  GLenum glLightID;
	  switch(id) {
	  case 0 : glLightID = GL_LIGHT0; break;
	  case 1 : glLightID = GL_LIGHT1; break;
	  case 2 : glLightID = GL_LIGHT2; break;
	  case 3 : glLightID = GL_LIGHT3; break;
	  case 4 : glLightID = GL_LIGHT4; break;
	  case 5 : glLightID = GL_LIGHT5; break;
	  case 6 : glLightID = GL_LIGHT6; break;
	  case 7 : glLightID = GL_LIGHT7; break;
	  default: glLightID = GL_LIGHT7; break;
	  }

	  // Set the last component of the position to 0 to indicate
	  // a directional light source
	  GLfloat glPosition[4];
	  glPosition[0] = -direction.x();
	  glPosition[1] = -direction.y();
	  glPosition[2] = -direction.z();
	  glPosition[3] = 0.0;

	  GLfloat glColor[4];
	  glColor[0] = color.r();
	  glColor[1] = color.g();
	  glColor[2] = color.b();
	  glColor[3] = 1.0;

	  glLightfv(glLightID, GL_POSITION, glPosition);
	  glLightfv(glLightID, GL_DIFFUSE, glColor);
	  glLightfv(glLightID, GL_SPECULAR, glColor);
	  glEnable(glLightID);
  }


private:

  // REPRESENTATION
  Vec3f direction;

};

// ====================================================================
class PointLight : public Light{

public:
	  PointLight(const Vec3f &p, const Vec3f &c){
		  position = p;
		  color = c;
	  }
	  virtual ~PointLight() {}
	  void getIllumination (const Vec3f &p, Vec3f &dir, Vec3f &col, float &distance) const {
	    dir = position-p;
	    distance = dir.Length();
	    dir.Normalize();
	    col = color; }

	  void print(){
		  cout << "Point Light: "
				<< " pos=(" << position.x() << "," << position.y() << "," << position.z() << ")"
				<< " color=(" << color.r() << "," << color.g() << "," << color.b() << ")" << endl;
	  }

	  // ====================================================================
	  // Create an OpenGL point light source with the appropriate
	  // color and position
	  // ====================================================================

	  void glInit(int id) {
		  GLenum glLightID;
		  switch(id) {
		  case 0 : glLightID = GL_LIGHT0; break;
		  case 1 : glLightID = GL_LIGHT1; break;
		  case 2 : glLightID = GL_LIGHT2; break;
		  case 3 : glLightID = GL_LIGHT3; break;
		  case 4 : glLightID = GL_LIGHT4; break;
		  case 5 : glLightID = GL_LIGHT5; break;
		  case 6 : glLightID = GL_LIGHT6; break;
		  case 7 : glLightID = GL_LIGHT7; break;
		  default: glLightID = GL_LIGHT7; break;
		  }

		  // Set the last component of the position to 1 to indicate
		  // a point light source
		  GLfloat glPosition[4];
		  glPosition[0] = position.x();
		  glPosition[1] = position.y() + 10;
		  glPosition[2] = position.z();
		  glPosition[3] = 1.0;

		  GLfloat glColor[4];
		  glColor[0] = color.r();
		  glColor[1] = color.g();
		  glColor[2] = color.b();
		  glColor[3] = 1.0;

		  glLightfv(glLightID, GL_POSITION, glPosition);
		  glLightfv(glLightID, GL_DIFFUSE, glColor);
		  glLightfv(glLightID, GL_SPECULAR, glColor);
		  glEnable(glLightID);
	  }

	  // ====================================================================
	  // ====================================================================
private:

	  Vec3f position;

};

// ====================================================================


#endif




